#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

"""
Methods to work with FABRIC `network services`_.

.. _`network services`: https://learn.fabric-testbed.net/knowledge-base/glossary/#network_service
"""

from __future__ import annotations

import logging
import threading
from typing import TYPE_CHECKING, List, Union

from fim.slivers.path_info import Path
from fim.user import ERO, Gateway
from tabulate import tabulate

if TYPE_CHECKING:
    from fabrictestbed_extensions.fablib.slice import Slice
    from fabrictestbed_extensions.fablib.interface import Interface
    from fabric_cf.orchestrator.swagger_client import Sliver as OrchestratorSliver

import ipaddress
import json
from ipaddress import IPv4Address, IPv4Network, IPv6Address, IPv6Network

import jinja2
from fabrictestbed.slice_editor import Capacities, Labels
from fabrictestbed.slice_editor import NetworkService as FimNetworkService
from fabrictestbed.slice_editor import ServiceType, UserData
from fim.slivers.network_service import NSLayer, ServiceType
from fim.user.network_service import MirrorDirection


class NetworkService:
    """
    A class for working with FABRIC network services.
    """

    network_service_map = {
        "L2Bridge": ServiceType.L2Bridge,
        "L2PTP": ServiceType.L2PTP,
        "L2STS": ServiceType.L2STS,
        "PortMirror": ServiceType.PortMirror,
        "FABNetv4": ServiceType.FABNetv4,
        "FABNetv6": ServiceType.FABNetv6,
        "FABNetv4Ext": ServiceType.FABNetv4Ext,
        "FABNetv6Ext": ServiceType.FABNetv6Ext,
        "L3VPN": ServiceType.L3VPN,
    }

    # Type names used in fim network services
    fim_l2network_service_types = ["L2Bridge", "L2PTP", "L2STS"]
    fim_l3network_service_types = [
        "FABNetv4",
        "FABNetv6",
        "FABNetv4Ext",
        "FABNetv6Ext",
        "L3VPN",
    ]

    fim_special_service_types = ["PortMirror"]

    @staticmethod
    def __get_fim_l2network_service_types() -> List[str]:
        """
        Not intended for API use. Returns a list of FIM L2 network service types.

        :return: List of FIM L2 network service types.
        :rtype: List[str]
        """
        return NetworkService.fim_l2network_service_types

    @staticmethod
    def __get_fim_l3network_service_types() -> List[str]:
        """
        Not intended for API use. Returns a list of FIM L3 network service types.

        :return: List of FIM L3 network service types.
        :rtype: List[str]
        """
        return NetworkService.fim_l3network_service_types

    @staticmethod
    def __get_fim_special_service_types() -> List[str]:
        """
        Not intended for API use. Returns a list of FIM special service types.

        :return: List of FIM special service types.
        :rtype: List[str]
        """
        return NetworkService.fim_special_service_types

    @staticmethod
    def get_fim_network_service_types() -> List[str]:
        """
        Not intended for API use. Returns a list of all FIM network service types.

        :return: List of all FIM network service types.
        :rtype: List[str]
        """
        return (
            NetworkService.__get_fim_l2network_service_types()
            + NetworkService.__get_fim_l3network_service_types()
            + NetworkService.__get_fim_special_service_types()
        )

    @staticmethod
    def __calculate_l2_nstype(
        interfaces: List[Interface] = None, ero_enabled: bool = False
    ) -> ServiceType:
        """
        Not inteded for API use

        Determines the L2 network service type based on the number of interfaces inputted.

        :param interfaces: a list of interfaces
        :type interfaces: list[Interface]

        :param ero_enabled: Flag indicating if ERO is specified
        :type ero_enabled: bool

        :raises Exception: if no network service type is not appropriate for the number of interfaces
        :return: the network service type
        :rtype: ServiceType
        """

        from fabrictestbed_extensions.fablib.facility_port import FacilityPort

        # if there is a basic NIC, WAN must be STS
        basic_nic_count = 0

        sites = set([])
        includes_facility_port = False
        facility_port_interfaces = 0
        for interface in interfaces:
            sites.add(interface.get_site())
            if isinstance(interface.get_node(), FacilityPort):
                includes_facility_port = True
                facility_port_interfaces += 1
            if interface.get_model() == "NIC_Basic":
                basic_nic_count += 1

        rtn_nstype = None
        if 1 >= len(sites) >= 0:
            rtn_nstype = NetworkService.network_service_map["L2Bridge"]
        # elif basic_nic_count == 0 and len(sites) == 2 and len(interfaces) == 2:
        #    #TODO: remove this when STS works on all links.
        #    rtn_nstype = NetworkService.network_service_map['L2PTP']
        elif len(sites) == 2:
            # Use L2STS when connecting two facility ports instead of L2PTP
            # L2PTP limitation for Facility Ports:
            # basically the layer-2 point-to-point server template applied is not popping
            # vlan tags over the MPLS tunnel between two facility ports.
            if (
                (includes_facility_port and facility_port_interfaces < 2) or ero_enabled
            ) and not basic_nic_count:
                # For now WAN FacilityPorts require L2PTP
                rtn_nstype = NetworkService.network_service_map["L2PTP"]
            elif len(interfaces) >= 2:
                rtn_nstype = NetworkService.network_service_map["L2STS"]
        else:
            raise Exception(
                f"Invalid Network Service: Networks are limited to 2 unique sites. Site requested: {sites}"
            )

        return rtn_nstype

    @staticmethod
    def __validate_nstype(type, interfaces):
        """
        Not intended for API use


        Verifies the network service type against the number of interfaces.

        :param type: the network service type to check
        :type type: ServiceType
        :param interfaces: the list of interfaces to check
        :type interfaces: list[Interface]
        :raises Exception: if the network service type is invalid based on the number of interfaces
        :return: true if the network service type is valid based on the number of interfaces
        :rtype: bool
        """
        # Just an empty network created; NS type would be validated when add_interface is invoked.
        if not len(interfaces):
            return True

        from fabrictestbed_extensions.fablib.facility_port import FacilityPort

        sites = set([])
        nics = set([])
        nodes = set([])
        for interface in interfaces:
            try:
                sites.add(interface.get_site())
                nics.add(interface.get_model())
                nodes.add(interface.get_node())
            except Exception as e:
                logging.info(
                    f"validate_nstype: skipping interface {interface.get_name()}, likely its a facility port"
                )

        # models: 'NIC_Basic', 'NIC_ConnectX_6', 'NIC_ConnectX_5'
        if type == NetworkService.network_service_map["L2Bridge"]:
            if len(sites) > 1:
                raise Exception(
                    f"Network type {type} must be empty or include interfaces from exactly one site. {len(sites)} "
                    f"sites requested: {sites}"
                )

        elif type == NetworkService.network_service_map["L2PTP"]:
            if not len(sites) == 2:
                raise Exception(
                    f"Network type {type} must include interfaces from exactly two sites. {len(sites)} sites "
                    f"requested: {sites}"
                )
            if "NIC_Basic" in nics:
                raise Exception(
                    f"Network type {type} does not support interfaces of type 'NIC_Basic'"
                )

        elif type == NetworkService.network_service_map["L2STS"]:
            exception_list = []
            if len(sites) != 2:
                exception_list.append(
                    f"Network type {type} must include interfaces from exactly two sites. {len(sites)} sites "
                    f"requested: {sites}"
                )
            if len(interfaces) > 2:
                hosts = set([])
                nodes_per_site = {}
                for interface in interfaces:
                    node = interface.get_node()
                    if node.get_site() not in nodes_per_site:
                        nodes_per_site[node.get_site()] = 0
                    if isinstance(node, FacilityPort):
                        continue
                    nodes_per_site[node.get_site()] += 1
                for interface in interfaces:
                    node = interface.get_node()
                    if (
                        interface.get_model() == "NIC_Basic"
                        and nodes_per_site[node.get_site()] > 1
                    ):
                        if node.get_host() is None:
                            exception_list.append(
                                f"Network type {type} does not support multiple NIC_Basic interfaces on VMs "
                                f"residing on the same host. Please see Node.set_host(host_name) to explicitly "
                                f"bind a nodes to a specific host. Node {node.get_name()} is unbound."
                            )
                        elif node.get_host() in hosts:
                            exception_list.append(
                                f"Network type {type} does not support multiple NIC_Basic interfaces on VMs residing "
                                f"on the same host. Please see Node.set_host(host_name) to explicitly bind a nodes "
                                f"to a specific host. Multiple nodes bound to {node.get_host()}."
                            )
                        else:
                            hosts.add(node.get_host())

            if len(exception_list) > 0:
                raise Exception(f"{exception_list}")
        else:
            raise Exception(f"Unknown network type {type}")

        return True

    @staticmethod
    def new_portmirror_service(
        slice: Slice = None,
        name: str = None,
        mirror_interface_name: str = None,
        mirror_interface_vlan: str = None,
        receive_interface: Interface or None = None,
        mirror_direction: str = "both",
    ) -> NetworkService:
        """
        Instantiate a new PortMirror service.

        ``mirror_direction`` can be ``"rx"``, ``"tx"`` or ``"both"``
        (non-case-sensitive)
        """
        # decode the direction
        if not isinstance(mirror_interface_name, str):
            raise Exception(
                f"When creating a PortMirror service mirror_interface is specified by name"
            )
        if not isinstance(mirror_direction, str):
            raise Exception(
                f'When creating a PortMirror service mirror_direction is a string "rx", "tx" or "both"'
                f'defaulting to "both"'
            )
        if not receive_interface:
            raise Exception(
                f"For PortMirror service the receiving interface must be specified upfront"
            )
        direction = MirrorDirection.Both
        # enable below when we are officially off python 3.9 and into 3.10 or higher
        # match mirror_direction.lower():
        #    case ['rx']:
        #        direction = MirrorDirection.RX_Only
        #    case ['tx']:
        #        direction = MirrorDirection.TX_Only
        #   case ['both']:
        #        direction = MirrorDirection.Both
        #   case _:
        #       raise Exception(f'Unknown direction specifier "{mirror_direction}" when creating PortMirror'
        #                        f'service {name}')
        no_case_direction = mirror_direction.lower()
        if no_case_direction == "rx":
            direction = MirrorDirection.RX_Only
        elif no_case_direction == "tx":
            direction = MirrorDirection.TX_Only
        elif no_case_direction == "both":
            pass
        else:
            raise Exception(
                f'Unknown direction specifier "{mirror_direction}" when creating PortMirror'
                f"service {name}"
            )
        logging.info(
            f"Create PortMirror Service: Slice: {slice.get_name()}, Network Name: {name} listening on "
            f"{mirror_interface_name} with direction {direction}"
        )
        fim_network_service = slice.topology.add_port_mirror_service(
            name=name,
            from_interface_name=mirror_interface_name,
            from_interface_vlan=mirror_interface_vlan,
            to_interface=receive_interface.fim_interface,
            direction=direction,
        )

        network_service = NetworkService(
            slice=slice, fim_network_service=fim_network_service
        )
        network_service.init_fablib_data()
        return network_service

    @staticmethod
    def new_l3network(
        slice: Slice = None,
        name: str = None,
        interfaces: List[Interface] = [],
        type: str = None,
        user_data={},
        technology: str = None,
        subnet: ipaddress.ip_network = None,
    ):
        """
        Not inteded for API use. See slice.add_l3network
        """
        if type == "IPv6":
            nstype = ServiceType.FABNetv6
        elif type == "IPv4":
            nstype = ServiceType.FABNetv4
        elif type == "IPv4Ext":
            nstype = ServiceType.FABNetv4Ext
        elif type == "IPv6Ext":
            nstype = ServiceType.FABNetv6Ext
        elif type == "L3VPN":
            nstype = ServiceType.L3VPN
        else:
            raise Exception(
                "Invalid L3 Network Type: Allowed values [IPv4, IPv4Ext, IPv6, IPv6Ext, L3VPN]"
            )

        # TODO: need a fabnet version of this
        # validate nstype and interface List
        # NetworkService.validate_nstype(nstype, interfaces)

        return NetworkService.__new_network_service(
            slice=slice,
            name=name,
            nstype=nstype,
            interfaces=interfaces,
            user_data=user_data,
            technology=technology,
            subnet=subnet,
        )

    @staticmethod
    def new_l2network(
        slice: Slice = None,
        name: str = None,
        interfaces: List[Interface] = [],
        type: str = None,
        user_data: dict = {},
    ):
        """
        Not inteded for API use. See slice.add_l2network

        Creates a new L2 network service.

        :param slice: the fablib slice to build this network on
        :type slice: Slice
        :param name: the name of the new network
        :type name: str
        :param interfaces: a list of interfaces to build the network service on
        :type interfaces: list[Interface]
        :param type: the type of network service to build (optional)
        :tyep type: str
        :return: the new L2 network service
        :rtype: NetworkService
        """
        if type is None:
            nstype = NetworkService.__calculate_l2_nstype(interfaces=interfaces)
        else:
            if type in NetworkService.__get_fim_l2network_service_types():
                nstype = NetworkService.network_service_map[type]
            else:
                raise Exception(
                    f"Invalid l2 network type: {type}. Please choose from "
                    f"{NetworkService.__get_fim_l2network_service_types()} or None for automatic selection"
                )

        # validate nstype and interface List
        NetworkService.__validate_nstype(nstype, interfaces)

        # Set default VLANs for P2P networks that did not pass in VLANs
        if nstype == ServiceType.L2PTP and len(
            interfaces
        ):  # or nstype == ServiceType.L2STS:
            vlan1 = interfaces[0].get_vlan()
            vlan2 = interfaces[1].get_vlan()

            if vlan1 is None and vlan2 is None:
                # TODO: Long term we might have multiple vlan on one property
                # and will need to make sure they are unique.  For now this okay
                interfaces[0].set_vlan("100")
                interfaces[1].set_vlan("100")
            elif vlan1 is None and vlan2 is not None:
                # Match VLANs if one is set.
                interfaces[0].set_vlan(vlan2)
            elif vlan1 is not None and vlan2 is None:
                # Match VLANs if one is set.
                interfaces[1].set_vlan(vlan1)

            # for interface in interfaces:
            #    if interface.get_model() != 'NIC_Basic' and not interface.get_vlan():
            #
            #        interface.set_vlan("100")
        network_service = NetworkService.__new_network_service(
            slice=slice,
            name=name,
            nstype=nstype,
            interfaces=interfaces,
            user_data=user_data,
        )
        return network_service

    @staticmethod
    def __new_network_service(
        slice: Slice = None,
        name: str = None,
        nstype: ServiceType = None,
        interfaces: List[Interface] = [],
        user_data: dict = {},
        technology: str = None,
        subnet: ipaddress.ip_network = None,
    ):
        """
        Not intended for API use. See slice.add_l2network


        Creates a new FABRIC network service and returns the fablib instance.

        :param slice: the fabric slice to build the network service with
        :type slice: Slice

        :param name: the name of the new network service
        :type name: str

        :param nstype: the type of network service to create
        :type nstype: ServiceType

        :param interfaces: a list of interfaces to
        :type interfaces: List

        :param technology: Specify the technology used should be set to AL2S when using for AL2S peering; otherwise None
        :type technology: str

        :param subnet: Request a specific subnet for FabNetv4, FabNetv6 or FabNetv6Ext services.
                       It's ignored for any other services.
        :type ipaddress.ip_network

        :return: the new fablib network service
        :rtype: NetworkService
        """
        fim_interfaces = []
        for interface in interfaces:
            fim_interfaces.append(interface.get_fim_interface())

        logging.info(
            f"Create Network Service: Slice: {slice.get_name()}, Network Name: {name}, Type: {nstype}"
        )
        fim_network_service = slice.topology.add_network_service(
            name=name, nstype=nstype, interfaces=fim_interfaces, technology=technology
        )

        if subnet:
            if nstype == ServiceType.FABNetv4:
                fim_network_service.gateway = Gateway(
                    lab=Labels(
                        ipv4_subnet=subnet.with_prefixlen,
                        ipv4=str(next(subnet.hosts())),
                    )
                )
            elif nstype in [ServiceType.FABNetv6, ServiceType.FABNetv6Ext]:
                fim_network_service.gateway = Gateway(
                    lab=Labels(
                        ipv6_subnet=subnet.with_prefixlen,
                        ipv6=str(next(subnet.hosts())),
                    )
                )

        network_service = NetworkService(
            slice=slice, fim_network_service=fim_network_service
        )
        network_service.set_user_data(user_data)
        network_service.init_fablib_data()

        return network_service

    @staticmethod
    def get_l3network_services(slice: Slice = None) -> list:
        """
        Gets all L3 networks services in this slice

        :return: List of all network services in this slice
        :rtype: List[NetworkService]
        """
        topology = slice.get_fim_topology()

        rtn_network_services = []
        fim_network_service = None
        logging.debug(
            f"NetworkService.get_fim_l3network_service_types(): {NetworkService.__get_fim_l3network_service_types()}"
        )

        for net_name, net in topology.network_services.items():
            logging.debug(f"scanning network: {net_name}, net: {net}")
            if (
                str(net.get_property("type"))
                in NetworkService.__get_fim_l3network_service_types()
            ):
                logging.debug(f"returning network: {net_name}, net: {net}")
                rtn_network_services.append(
                    NetworkService(slice=slice, fim_network_service=net)
                )

        return rtn_network_services

    @staticmethod
    def get_l3network_service(slice: Slice = None, name: str = None):
        """
        Gets a particular L3 network service from this slice.


        :param name: Name network
        :type name: String
        :return: network services on this slice
        :rtype: list[NetworkService]
        """
        for net in NetworkService.get_l3network_services(slice=slice):
            if net.get_name() == name:
                return net

        raise Exception(f"Network not found. Slice {slice.slice_name}, network {name}")

    @staticmethod
    def get_l2network_services(slice: Slice = None) -> list:
        """
        Not inteded for API use.

        Gets a list of L2 network services on a fablib slice.

        :param slice: the fablib slice from which to get the network services
        :type slice: Slice
        :return: a list of network services on slice
        :rtype: list[NetworkService]
        """
        topology = slice.get_fim_topology()

        rtn_network_services = []
        fim_network_service = None
        for net_name, net in topology.network_services.items():
            if (
                str(net.get_property("type"))
                in NetworkService.__get_fim_l2network_service_types()
            ):
                rtn_network_services.append(
                    NetworkService(slice=slice, fim_network_service=net)
                )

        return rtn_network_services

    @staticmethod
    def get_l2network_service(slice: Slice = None, name: str = None):
        """
        Not inteded for API use.

        Gets a particular network service on a fablib slice.

        :param slice: the fablib slice from which to get the network service
        :type slice: Slice
        :param name: the name of the network service to get
        :type name: str
        :raises Exception: if the network is not found
        :return: the particular network service
        :rtype: NetworkService
        """
        for net in NetworkService.get_l2network_services(slice=slice):
            if net.get_name() == name:
                return net

        raise Exception(f"Network not found. Slice {slice.slice_name}, network {name}")

    @staticmethod
    def get_network_services(slice: Slice = None) -> list:
        """
        Gets all network services (L2 and L3) in this slice

        :return: List of all network services in this slice
        :rtype: List[NetworkService]
        """

        topology = slice.get_fim_topology()

        rtn_network_services = []
        fim_network_service = None
        for net_name, net in topology.network_services.items():
            if (
                str(net.get_property("type"))
                in NetworkService.get_fim_network_service_types()
            ):
                rtn_network_services.append(
                    NetworkService(slice=slice, fim_network_service=net)
                )

        return rtn_network_services

    @staticmethod
    def get_network_service(slice: Slice = None, name: str = None):
        """
        Gest a particular network service from this slice.

        :param name: the name of the network service to search for
        :type name: str
        :return: a particular network service
        :rtype: NetworkService
        """
        for net in NetworkService.get_network_services(slice=slice):
            if net.get_name() == name:
                return net

        raise Exception(f"Network not found. Slice {slice.slice_name}, network {name}")

    def __init__(
        self, slice: Slice = None, fim_network_service: FimNetworkService = None
    ):
        """
        .. note::

            Not inteded for API use.

        :param slice: the fablib slice to set as instance state
        :type slice: Slice

        :param fim_network_service: the FIM network service to set as
            instance state
        :type fim_network_service: FimNetworkService
        """
        super().__init__()
        self.fim_network_service = fim_network_service
        self.slice = slice

        self.interfaces = None

        try:
            if self.slice.isStable():
                self.sliver = self.slice.get_sliver(
                    reservation_id=self.get_reservation_id()
                )
        except:
            pass

        self.sliver = None
        self.lock = threading.Lock()

    def __str__(self):
        """
        Creates a tabulated string describing the properties of the network service.

        Intended for printing network service information.

        :return: Tabulated string of network service information
        :rtype: String
        """
        table = [
            ["ID", self.get_reservation_id()],
            ["Name", self.get_name()],
            ["Layer", self.get_layer()],
            ["Type", self.get_type()],
            ["Site", self.get_site()],
            ["Gateway", self.get_gateway()],
            ["Subnet", self.get_subnet()],
            ["State", self.get_reservation_state()],
            ["Error", self.get_error_message()],
        ]

        return tabulate(table)  # , headers=["Property", "Value"])

    def toJson(self):
        """
        Returns the network attributes as a json string

        :return: network attributes as json string
        :rtype: str
        """
        return json.dumps(self.toDict(), indent=4)

    @staticmethod
    def get_pretty_name_dict():
        """
        Return mappings from non-pretty names to pretty names.

        Pretty names are used when rendering table headers.
        """
        return {
            "id": "ID",
            "name": "Name",
            "layer": "Layer",
            "type": "Type",
            "site": "Site",
            "gateway": "Gateway",
            "subnet": "Subnet",
            "state": "State",
            "error": "Error",
        }

    def toDict(self, skip=[]):
        """
        Returns the network attributes as a dictionary

        :return: network attributes as dictionary
        :rtype: dict
        """
        return {
            "id": str(self.get_reservation_id()),
            "name": str(self.get_name()),
            "layer": str(self.get_layer()),
            "type": str(self.get_type()),
            "site": str(self.get_site()),
            "subnet": str(self.get_subnet()),
            "gateway": str(self.get_gateway()),
            "state": str(self.get_reservation_state()),
            "error": str(self.get_error_message()),
        }

    def generate_template_context(self):
        context = self.toDict()
        context["interfaces"] = []
        # for interface in self.get_interfaces():
        #    context["interfaces"].append(interface.get_name())

        return context

    def get_template_context(self):
        return self.get_slice().get_template_context(self)

    def render_template(self, input_string):
        environment = jinja2.Environment()
        environment.json_encoder = json.JSONEncoder(ensure_ascii=False)

        template = environment.from_string(input_string)
        output_string = template.render(self.get_template_context())

        return output_string

    def show(
        self, fields=None, output=None, quiet=False, colors=False, pretty_names=True
    ):
        """
        Show a table containing the current network attributes.

        There are several output options: "text", "pandas", and "json"
        that determine the format of the output that is returned and
        (optionally) displayed/printed.  This is controlled by the
        parameter ``output``, which can be one of:

        - ``"text"``: string formatted with tabular
        - ``"pandas"``: pandas dataframe
        - ``"json"``: string in json format

        :param output: Output format (``"text"``, ``"pandas"``, or
            ``"json"``)
        :type output: str

        :param fields: List of fields to show.  JSON output will
            include all available fields.  Example:
            ``fields=['Name','State']``
        :type fields: List[str]

        :param quiet: True to specify printing/display
        :type quiet: bool

        :param colors: True to specify state colors for pandas output
        :type colors: bool

        :return: table in format specified by output parameter
        :rtype: Object
        """

        data = self.toDict()

        # fields = ["ID", "Name", "Layer", "Type", "Site",
        #        "Gateway", "Subnet","State", "Error",
        #         ]

        if pretty_names:
            pretty_names_dict = self.get_pretty_name_dict()
        else:
            pretty_names_dict = {}

        node_table = self.get_fablib_manager().show_table(
            data,
            fields=fields,
            title="Network",
            output=output,
            quiet=quiet,
            pretty_names_dict=pretty_names_dict,
        )

        return node_table

    def get_slice(self) -> Slice:
        """
        Gets the fablib slice this network service is built on.

        :return: the slice this network is on
        :rtype: Slice
        """
        return self.slice

    def get_fablib_manager(self):
        """
        Get a reference to :py:class:`.FablibManager`.
        """
        return self.get_slice().get_fablib_manager()

    def get_site(self) -> str or None:
        """
        Gets site name on network service.
        """
        try:
            return self.get_sliver().fim_sliver.site
        except Exception as e:
            logging.warning(f"Failed to get site: {e}")

            return None

    def get_layer(self) -> str or None:
        """
        Gets the layer of the network services (L2 or L3)

        :return: L2 or L3
        :rtype: String
        """
        try:
            return self.get_fim().get_property(pname="layer")
            # return self.get_sliver().fim_sliver.layer
        except Exception as e:
            logging.warning(f"Failed to get layer: {e}")
            return None

    def get_type(self):
        """
        Gets the type of the network services

        :return: network service types
        :rtype: String
        """
        try:
            return self.get_fim().type
            # return self.get_sliver().fim_sliver.resource_type
        except Exception as e:
            logging.warning(f"Failed to get type: {e}")
            return None

    def get_sliver(self) -> OrchestratorSliver:
        """
        Gets the sliver.
        """
        if not self.sliver and self.slice.isStable():
            self.sliver = self.slice.get_sliver(
                reservation_id=self.get_reservation_id()
            )
        return self.sliver

    def get_fim_network_service(self) -> FimNetworkService:
        """
        Not recommended for most users.

        Gets the FABRIC network service this instance represents.

        :return: the FIM network service
        :rtype: FIMNetworkService
        """
        return self.fim_network_service

    def get_error_message(self) -> str:
        """
        Gets the error messages

        :return: network service types
        :rtype: String
        """
        try:
            return (
                self.get_fim_network_service()
                .get_property(pname="reservation_info")
                .error_message
            )
        except:
            return ""

    def get_gateway(self) -> IPv4Address or IPv6Address or None:
        """
        Gets the assigend gateway for a FABnetv L3 IPv6 or IPv4 network

        :return: gateway IP
        :rtype: IPv4Address or IPv6Network
        """
        try:
            gateway = None

            if self.is_instantiated():
                if self.get_layer() == NSLayer.L3:
                    gateway = ipaddress.ip_address(
                        self.get_sliver().fim_sliver.gateway.gateway
                    )
                    # if self.get_type() in [ServiceType.FABNetv6, ServiceType.FABNetv6Ext]:
                #    gateway = IPv6Address(self.get_sliver().fim_sliver.gateway.gateway)
                # elif self.get_type() in [ServiceType.FABNetv4, ServiceType.FABNetv4Ext]:
                #    gateway = IPv4Address(self.get_sliver().fim_sliver.gateway.gateway)
                else:
                    # L2 Network
                    fablib_data = self.get_fablib_data()
                    try:
                        gateway = ipaddress.ip_address(fablib_data["subnet"]["gateway"])
                    except Exception as e:
                        gateway = None
            else:
                gateway = f"{self.get_name()}.gateway"

            return gateway
        except Exception as e:
            logging.warning(f"Failed to get gateway: {e}")

    def get_available_ips(
        self, count: int = 256
    ) -> List[IPv4Address or IPv6Address] or None:
        """
        Gets the IPs available for a FABnet L3 network

        Note: large IPv6 address spaces take considerable time to build this
        list. By default this will return the first 256 addresses.  If you needed
        more addresses, set the count parameter.

        :param count: number of addresse to include
        :type slice: Slice
        :return: gateway IP
        :rtype: List[IPv4Address]
        """
        try:
            ip_list = []
            gateway = self.get_gateway()
            for i in range(count):
                logging.debug(f"adding IP {i}")
                ip_list.append(gateway + i + 1)
            return ip_list
        except Exception as e:
            logging.warning(f"Failed to get_available_ips: {e}")

    def get_public_ips(self) -> Union[List[IPv4Address] or List[IPv6Address] or None]:
        """
        Get list of public IPs assigned to the FabNetv*Ext service
        :return: List of Public IPs
        :rtype: List[IPv4Address] or List[IPv6Address] or None
        """
        if self.get_fim_network_service().labels is not None:
            if self.get_fim_network_service().labels.ipv4 is not None:
                result = []
                for x in self.get_fim_network_service().labels.ipv4:
                    result.append(IPv4Address(x))
                return result
            elif self.get_fim_network_service().labels.ipv6 is not None:
                result = []
                for x in self.get_fim_network_service().labels.ipv6:
                    result.append(IPv6Address(x))
                return result
        return None

    def get_subnet(self) -> IPv4Network or IPv6Network or None:
        """
        Gets the assigned subnet for a FABnet L3 IPv6 or IPv4 network

        :return: gateway IP
        :rtype: IPv4Network or IPv6Network
        """
        try:
            subnet = None
            if self.is_instantiated():
                if self.get_layer() == NSLayer.L3:
                    subnet = ipaddress.ip_network(
                        self.get_sliver().fim_sliver.gateway.subnet
                    )
                else:
                    # L2 Network
                    fablib_data = self.get_fablib_data()
                    try:
                        subnet = ipaddress.ip_network(fablib_data["subnet"]["subnet"])
                    except Exception as e:
                        logging.warning(f"Failed to get L2 subnet: {e}")

                        subnet = None
            else:
                subnet = f"{self.get_name()}.subnet"

            return subnet
        except Exception as e:
            logging.warning(f"Failed to get subnet: {e}")

    def get_reservation_id(self) -> str or None:
        """
        Gets the reservation id of the network

        :return: reservation ID
        :rtype: String
        """
        try:
            return (
                self.get_fim_network_service()
                .get_property(pname="reservation_info")
                .reservation_id
            )
        except Exception as e:
            logging.debug(f"Failed to get reservation_id: {e}")
            return None

    def get_reservation_state(self) -> str or None:
        """
        Gets the reservation state of the network

        :return: reservation state
        :rtype: String
        """
        try:
            return (
                self.get_fim_network_service()
                .get_property(pname="reservation_info")
                .reservation_state
            )
        except Exception as e:
            logging.warning(f"Failed to get reservation_state: {e}")
            return None

    def get_name(self) -> str:
        """
        Gets the name of this network service.

        :return: the name of this network service
        :rtype: String
        """
        return self.get_fim_network_service().name

    def get_interfaces(self) -> List[Interface]:
        """
        Gets the interfaces on this network service.

        :return: the interfaces on this network service
        :rtype: List[Interfaces]
        """
        if not self.interfaces:
            self.interfaces = []
            for interface in self.get_fim_network_service().interface_list:
                logging.debug(f"interface: {interface}")

                try:
                    self.interfaces.append(
                        self.get_slice().get_interface(name=interface.name)
                    )
                except:
                    logging.warning(f"interface not found: {interface.name}")
                    """ Commenting this code as not sure why this was added for now.
                    from fabrictestbed_extensions.fablib.interface import Interface

                    self.interfaces.append(
                        Interface(fim_interface=interface, node=self)
                    )
                    """

        return self.interfaces

    def get_interface(self, name: str = None) -> Interface or None:
        """
        Gets a particular interface on this network service.

        :param name: the name of the interface to search for
        :type name: str
        :return: the particular interface
        :rtype: Interface
        """
        for interface in self.get_fim_network_service().interface_list:
            if name in interface.name:
                return self.get_slice().get_interface(name=interface.name)

    def has_interface(self, interface: Interface) -> bool:
        """
        Determines whether this network service has a particular interface.

        :param interface: the fablib interface to search for
        :type interface: Interface
        :return: whether this network service has interface
        :rtype: bool
        """
        for fim_interface in self.get_fim_network_service().interface_list:
            # print(f"fim_interface.name: {fim_interface.name}, interface.get_name(): {interface.get_name()}")
            if fim_interface.name.endswith(interface.get_name()):
                return True

        return False

    def get_fim(self):
        """
        Gets the FABRIC information model (FIM) object.
        """
        return self.get_fim_network_service()

    def set_user_data(self, user_data: dict):
        """
        Set user data.

        :param user_data: a `dict`.
        """
        self.get_fim().set_property(
            pname="user_data", pval=UserData(json.dumps(user_data))
        )

    def get_user_data(self):
        """
        Get user data.
        """
        try:
            return json.loads(str(self.get_fim().get_property(pname="user_data")))
        except:
            return {}

    def __replace_network_service(self, nstype: ServiceType):
        fim_interfaces = []
        name = self.get_name()

        for interface in self.get_interfaces():
            fim_interfaces.append(interface.get_fim_interface())
            self.get_fim().disconnect_interface(interface=interface.get_fim())

        user_data = self.get_user_data()
        self.get_slice().topology.remove_network_service(name=self.get_name())

        self.fim_network_service = self.get_slice().topology.add_network_service(
            name=name, nstype=nstype, interfaces=fim_interfaces
        )
        self.set_user_data(user_data)

    def add_interface(self, interface: Interface):
        """
        Add an :py:class:`.Interface` to the network service.
        """
        if self.get_type() == ServiceType.PortMirror:
            raise Exception(
                "Interfaces cannot be attached to PortMirror service - they can only"
                "be specified at service creation"
            )

        iface_fablib_data = interface.get_fablib_data()

        new_interfaces = self.get_interfaces()
        new_interfaces.append(interface)

        curr_nstype = self.get_type()
        if self.get_layer() == NSLayer.L2:
            ero_enabled = True if self.get_fim().ero else False
            new_nstype = NetworkService.__calculate_l2_nstype(
                interfaces=new_interfaces, ero_enabled=ero_enabled
            )
            if curr_nstype != new_nstype:
                self.__replace_network_service(new_nstype)
            else:
                self.get_fim().connect_interface(interface=interface.get_fim())
        elif self.get_layer() == NSLayer.L3 and self.is_instantiated():
            if interface.get_site() != self.get_site():
                raise Exception("L3 networks can only include nodes from one site")

        if "addr" in iface_fablib_data:
            addr = iface_fablib_data["addr"]
        else:
            addr = None

        if "auto" in iface_fablib_data:
            auto = iface_fablib_data["auto"]
        else:
            auto = False

        if self.get_subnet():
            if addr:
                iface_fablib_data["addr"] = str(self.allocate_ip(addr))
            elif auto:
                iface_fablib_data["addr"] = str(self.allocate_ip())

        interface.set_fablib_data(iface_fablib_data)

        if self.get_layer() == NSLayer.L3:
            self.get_fim().connect_interface(interface=interface.get_fim())

    def remove_interface(self, interface: Interface):
        """
        Remove an :py:class:`.Interface` from the network service.
        """
        iface_fablib_data = interface.get_fablib_data()

        self.free_ip(interface.get_ip_addr())

        interfaces = self.get_interfaces()

        if interface in interfaces:
            interfaces.remove(interface)

        curr_nstype = self.get_type()
        if self.get_layer() == NSLayer.L2:
            ero_enabled = True if self.get_fim().ero else False
            new_nstype = NetworkService.__calculate_l2_nstype(
                interfaces=interfaces, ero_enabled=ero_enabled
            )
            if curr_nstype != new_nstype:
                self.__replace_network_service(new_nstype)

        interface.set_fablib_data(iface_fablib_data)

        self.get_fim().disconnect_interface(interface=interface.get_fim())

    def delete(self):
        """
        Delete the network service.
        """
        self.get_slice().get_fim_topology().remove_network_service(name=self.get_name())

    def get_fablib_data(self):
        """
        Get value associated with `fablib_data` key of user data.
        """
        try:
            return self.get_user_data()["fablib_data"]
        except:
            return {}

    def set_fablib_data(self, fablib_data: dict):
        """
        Set value associated with `fablib_data` key of user data.
        """
        user_data = self.get_user_data()
        user_data["fablib_data"] = fablib_data
        self.set_user_data(user_data)

    def set_subnet(self, subnet: IPv4Network or IPv6Network):
        """
        Add subnet info to the network service.
        """
        fablib_data = self.get_fablib_data()
        if "subnet" not in fablib_data:
            fablib_data["subnet"] = {}
        fablib_data["subnet"]["subnet"] = str(subnet)
        fablib_data["subnet"]["allocated_ips"] = []
        self.set_fablib_data(fablib_data)

    def set_gateway(self, gateway: IPv4Address or IPv6Address):
        """
        Add gateway info to the network service.
        """
        fablib_data = self.get_fablib_data()
        if "subnet" not in fablib_data:
            fablib_data["subnet"] = {}
        fablib_data["subnet"]["gateway"] = str(gateway)
        self.set_fablib_data(fablib_data)

    def get_allocated_ips(self):
        """
        Get the list of IP addesses allocated for the network service.
        """
        try:
            allocated_ips = []
            for addr in self.get_fablib_data()["subnet"]["allocated_ips"]:
                allocated_ips.append(ipaddress.ip_address(addr))

            return allocated_ips
        except Exception as e:
            return []

    def set_allocated_ip(self, addr: IPv4Address or IPv6Address = None):
        """
        Add ``addr`` to the list of allocated IPs.
        """
        fablib_data = self.get_fablib_data()
        if "subnet" not in fablib_data:
            fablib_data["subnet"] = {}
        allocated_ips = fablib_data["subnet"]["allocated_ips"]
        allocated_ips.append(str(addr))
        self.set_fablib_data(fablib_data)

    def allocate_ip(self, addr: IPv4Address or IPv6Address = None):
        """
        Allocate an IP for the network service.
        """
        try:
            self.lock.acquire()
            subnet = self.get_subnet()
            allocated_ips = self.get_allocated_ips()

            if addr:
                # if addr != subnet.network_address and addr not in allocated_ips:
                if addr not in allocated_ips:
                    self.set_allocated_ip(addr)
                    return addr
            elif (
                type(subnet) == ipaddress.IPv4Network
                or type(subnet) == ipaddress.IPv6Network
            ):
                for host in subnet:
                    if host != subnet.network_address and host not in allocated_ips:
                        self.set_allocated_ip(host)

                        return host
            return None
        finally:
            self.lock.release()

    def set_allocated_ips(self, allocated_ips: list[IPv4Address or IPv6Address]):
        """
        Set a list of IPs to be "allocated IPs".
        """
        fablib_data = self.get_fablib_data()
        allocated_ips_strs = []
        for ip in allocated_ips:
            allocated_ips_strs.append(str(ip))

        if "subnet" not in fablib_data:
            fablib_data["subnet"] = {}

        fablib_data["subnet"]["allocated_ips"] = allocated_ips_strs
        self.set_fablib_data(fablib_data)

    def free_ip(self, addr: IPv4Address or IPv6Address):
        """
        Remove an IP from the list of allocated IPs.
        """
        try:
            self.lock.acquire()
            allocated_ips = self.get_allocated_ips()
            if addr in allocated_ips:
                allocated_ips.remove(addr)
            self.set_allocated_ips(allocated_ips)
        finally:
            self.lock.release()

    def make_ip_publicly_routable(self, ipv6: list[str] = None, ipv4: list[str] = None):
        """
        Mark a list of IPs as publicly routable.
        """
        labels = self.fim_network_service.labels
        if labels is None:
            labels = Labels()
        if self.fim_network_service.type == ServiceType.FABNetv4Ext:
            labels = Labels.update(labels, ipv4=ipv4)

        elif self.fim_network_service.type == ServiceType.FABNetv6Ext:
            labels = Labels.update(labels, ipv6=ipv6)

        self.fim_network_service.set_properties(labels=labels)

    def init_fablib_data(self):
        """
        Initialize fablib data.
        """
        fablib_data = {"instantiated": "False", "mode": "manual"}
        self.set_fablib_data(fablib_data)

    def is_instantiated(self):
        """
        Return ``True`` if network service has been instantiated.
        """
        fablib_data = self.get_fablib_data()
        if "instantiated" in fablib_data and fablib_data["instantiated"] == "True":
            return True
        else:
            return False

    def set_instantiated(self, instantiated: bool = True):
        """
        Set instantiated flag in the fablib_data saved in UserData
        blob in the FIM model.

        :param instantiated: flag indicating if the service has been
            instantiated or not
        :type instantiated: bool
        """
        fablib_data = self.get_fablib_data()
        fablib_data["instantiated"] = str(instantiated)
        self.set_fablib_data(fablib_data)

    def config(self):
        """
        Sets up the meta data for the Network Service

            - For layer3 services, Subnet, gateway and allocated IPs
              are updated/maintained fablib_data saved in UserData
              blob in the FIM model

            - For layer2 services, no action is taken
        """
        if not self.is_instantiated():
            self.set_instantiated(True)

        # init
        if self.get_layer() == NSLayer.L3:
            # init fablib data for fabnet networks
            self.set_subnet(self.get_subnet())
            self.set_gateway(self.get_gateway())
            allocated_ips = self.get_allocated_ips()
            if not allocated_ips:
                allocated_ips = []
            if self.get_gateway() not in allocated_ips:
                allocated_ips.append(self.get_gateway())
            self.set_allocated_ip(self.get_gateway())

    def peer(
        self,
        other: NetworkService,
        labels: Labels,
        peer_labels: Labels,
        capacities: Capacities,
    ):
        """
        Peer a network service; used for AL2S peering between FABRIC Networks and Cloud Networks
        Peer this network service to another. A few constraints are enforced like services being
        of the same type. Both services will have ServicePort interfaces facing each other over a link.
        It typically requires labels and capacities to put on the interface facing the other service

        :param other: network service to be peered
        :type other: NetworkService
        :param labels: labels
        :type labels: Labels
        :param peer_labels: peer labels
        :type peer_labels: Labels
        :param capacities: capacities
        :type capacities: Capacities

        """
        # Peer Cloud L3VPN with FABRIC L3VPN
        self.get_fim().peer(
            other.get_fim(),
            labels=labels,
            peer_labels=peer_labels,
            capacities=capacities,
        )

    def set_l2_route_hops(self, hops: List[str]):
        """
        Explicitly define the sequence of sites or hops to be used for a layer 2 connection.

        Users provide a list of site names, which are then mapped by the ControlFramework to the corresponding
        layer 2 loopback IP addresses utilized by the Explicit Route Options in the Network Service configuration
        on the switch.

        :param hops: A list of site names to be used as hops.
        :type hops: List[str]
        """
        # Do nothing if hops is None or empty list
        if not hops or not len(hops):
            return

        interfaces = self.get_interfaces()

        if len(interfaces) != 2 or self.get_type() not in [
            ServiceType.L2STS,
            ServiceType.L2PTP,
        ]:
            raise Exception(
                "Network path can only be specified for a Point to Point Layer2 connection!"
            )

        ifs_sites = []
        for ifs in interfaces:
            ifs_sites.append(ifs.get_site())

        resources = self.get_fablib_manager().get_resources()
        resources.validate_requested_ero_path(
            source=ifs_sites[0], end=ifs_sites[1], hops=hops
        )
        p = Path()
        p.set_symmetric(hops)
        e = ERO()
        e.set(payload=p)
        ns_type = self.__calculate_l2_nstype(interfaces=interfaces, ero_enabled=True)
        self.get_fim().set_properties(type=ns_type, ero=e)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Komal Thareja(kthare10@renci.org)
import json


class Artifact:
    pretty_names = {
        "title": "Title",
        "uuid": "ID",
        "description_short": "Short Description",
        "description_long": "Long Description",
        "project_name": "Project Name",
        "authors": "Authors",
        "versions": "Versions",
        "created": "Created",
        "modified": "Modified",
        "tags": "Tags",
    }

    def __init__(self, artifact_info: dict, fablib_manager):
        """
        Initialize a Artifact object.

        :param artifact_info: artifact_info
        :type artifact_info: dict

        :param fablib_manager: The manager for the Fabric library.
        :type fablib_manager: Any

        :return: None
        """
        self.fablib_manager = fablib_manager
        self.artifact_info = artifact_info

    def to_json(self) -> str:
        """
        Convert the Site object to a JSON string.

        :return: JSON string representation of the Site object.
        :rtype: str
        """
        return json.dumps(self.to_dict(), indent=4)

    def get_fablib_manager(self):
        """
        Get the Fabric library manager associated with the site.

        :return: The Fabric library manager.
        :rtype: Any
        """
        return self.fablib_manager

    def __str__(self) -> str:
        """
        Convert the Site object to a string representation in JSON format.

        :return: JSON string representation of the Site object.
        :rtype: str
        """
        return self.to_json()

    def to_dict(self) -> dict:
        """
        Convert artifact information into a dictionary
        """
        authors = []
        for a in self.artifact_info.get("authors"):
            authors.append(a.get("email"))
        versions = []
        for v in self.artifact_info.get("versions"):
            versions.append(f"{v.get('version')}#{v.get('urn')}")
        d = {
            "title": self.artifact_info.get("title"),
            "uuid": self.artifact_info.get("uuid"),
            "description_short": self.artifact_info.get("description_short"),
            # "description_long": self.artifact_info.get("description_long"),
            "project_name": self.artifact_info.get("project_name"),
            "authors": ", ".join(authors),
            "versions": "\n".join(versions),
            "visibility": self.artifact_info.get("visibility"),
            "created": self.artifact_info.get("created"),
            "modified": self.artifact_info.get("modified"),
        }

        return d



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)
import os
import traceback
from abc import ABC, abstractmethod
from typing import List

from fabrictestbed.slice_editor import (
    Capacities,
    ComponentCatalog,
    ComponentType,
    ExperimentTopology,
)
from fabrictestbed.slice_manager import SliceManager, SliceState, Status


class AbcTopologyEditor(ABC):
    EXPERIMENT_STATE_UNSUBMITTED = "unsubmitted"
    EXPERIMENT_STATE_LIVE = "live"
    EXPERIMENT_STATE_ERROR = "error"
    EXPERIMENT_STATE_DELETED = "deleted or error"

    DEFAULT_SLICE_SELECT_VALUE = "<Choose Slice>"

    def __init__(self):
        self.slice_manager = SliceManager()
        self.advertised_topology = None
        self.experiments = []
        self.current_experiment = None
        self.site_detail = False
        self.ssh_key = None
        with open(os.environ["HOME"] + "/.ssh/id_rsa.pub", "r") as fd:
            self.ssh_key = fd.read().strip()

        self.pull_advertised_topology()
        self.update_experiment_list()

    def pull_advertised_topology(self):
        return_status, self.advertised_topology = self.slice_manager.resources()
        if return_status != Status.OK:
            print(
                "Failed to get advertised_topology: {}".format(self.advertised_topology)
            )

    def remove_experiment(self, experiment):
        """
        Remove and experiment from the editor.
        Does not delete the slice.
        :param experiment:
        :return:
        """
        # Unset current experiment
        if experiment == self.current_experiment:
            self.current_experiment = None

        # remove the experiment
        self.experiments.remove(experiment)

    def delete_experiment(self, experiment):
        print("Delete Slice")
        if experiment["editor_slice_state"] != self.EXPERIMENT_STATE_UNSUBMITTED:
            try:
                print("self.current_experiment: " + str(self.current_experiment))
                print("self.experiments: " + str(self.experiments))

                # delete the slice
                result = self.slice_manager.delete(slice_object=experiment["slice"])

                experiment["editor_slice_state"] = self.EXPERIMENT_STATE_DELETED

                # TODO eliminate race condition... need to update until delete instead of sleep
                import time

                time.sleep(1)

            except Exception as e:
                # TODO: Should create popup or other user facing error message
                print("Failed to delete slice. Error: " + str(e))
                traceback.print_exc()

        # Remove experiment from editor
        self.remove_experiment(experiment)

    def create_experiment(self, slice_name):
        print("Create Experiment")

        experiment = self.get_experiment_by_name(slice_name)

        if experiment != None:
            # TODO: should raise Exception
            print(
                "Failed to create new slice. Error: Name not unique or invalid "
                + str(slice_name)
            )
            return

        # New experiment
        self.experiments.append(
            {
                "slice_name": slice_name,
                "editor_slice_state": self.EXPERIMENT_STATE_UNSUBMITTED,
                "topology": ExperimentTopology(),
                # ssh_keys
                # detail_levels {'detail': HIGH, {'node1': LOW, 'node2': MED}}
            }
        )

    def set_current_experiment(self, experiment):
        self.current_experiment = experiment

    def add_node(self, node_name):
        print("abc.add_node  node_name {}".format(node_name))

        # Add to FABRIC experiment topology
        new_node = self.current_experiment["topology"].add_node(
            name=node_name, site=self.DEFAULT_NODE_SITE_VALUE
        )
        cap = Capacities(
            core=self.DEFAULT_NODE_CORE_VALUE,
            ram=self.DEFAULT_NODE_RAM_VALUE,
            disk=self.DEFAULT_NODE_DISK_VALUE,
        )
        new_node.set_properties(
            capacities=cap,
            image_type=self.DEFAULT_NODE_IMAGE_TYPE_VALUE,
            image_ref=self.DEFAULT_NODE_IMAGE_VALUE,
        )

    def delete_node(self, node_name):
        """
        Delete node
        :param node_name: node name
        :return:
        """
        print("Delete Node")
        try:
            # Delete the node from experiment
            self.current_experiment["topology"].remove_node(node_name)
        except Exception as e:
            # TODO: Should create popup or other user facing error message
            print("Failed to delete node. Error: " + str(e))
            traceback.print_exc()

    def save_node(
        self,
        topology_node,
        experiment=None,
        node_name=None,
        site_name=None,
        cores=None,
        ram=None,
        disk=None,
        image=None,
        image_type=None,
    ):
        """
        Save node
        :param node_name:
        :param site_name:
        :param cores:
        :param ram:
        :param disk:
        :param image:
        :param image_type:
        :return:
        """
        print("ABC.save_node")
        print("save_node: topology_node {}".format(topology_node))
        print("save_node: experiment {}".format(experiment))
        print("save_node:  node_name {}".format(node_name))
        print("save_node:  site_name {}".format(site_name))
        print("save_node:  cores {}".format(cores))
        print("save_node:  ram {}".format(ram))
        print("save_node:  disk {}".format(disk))
        print("save_node:  image {}".format(image))
        print("save_node:  image_type {}".format(image_type))

        if experiment == None:
            experiment = self.current_experiment

        try:
            # node = experiment['topology'].nodes[node_name]

            # Save node properties to FABRIC topology
            if node_name:
                topology_node.set_property(pname="name", pval=node_name)
            if site_name:
                topology_node.set_property(pname="site", pval=site_name)

            # Set capacities
            cap = Capacities(core=cores, ram=ram, disk=disk)
            topology_node.set_properties(
                capacities=cap, image_type=image_type, image_ref=image
            )
        except Exception as e:
            # TODO: Should create popup or other user facing error message
            print("Failed to save node. Error: " + str(e))
            traceback.print_exc()

    def update_capacities(self):
        """
        Update capacities from the available resources information
        :return:
        """
        # CREATE CURRENT CAPACITIES DICT
        print("Update Capacities")
        capacities = {}

        for t in self.advertised_topology.sites.values():
            # name
            site_name = t.name
            # available capacities total vs. currently allocated ones
            available_capacities = t.get_property("capacities")
            allocated_capacities = t.get_property("capacity_allocations")

            if allocated_capacities is None:
                allocated_capacities = Capacities()

            if available_capacities is None:
                available_capacities = Capacities()

            d1 = available_capacities.__dict__.copy()
            d2 = allocated_capacities.__dict__.copy()

            # remove capacities that are not available
            for d in allocated_capacities.__dict__:
                if d1[d] == 0 and d2[d] == 0:
                    d1.pop(d)
                    d2.pop(d)

            # add available capacities to site
            site_capacities = {}
            for c in d1:
                site_capacities[c] = d1[c] - d2[c]

            capacities[site_name] = site_capacities

        return capacities

    def add_component(self, component_name, model_type, node=None):
        if node == None:
            node = self.current_node
        if component_name == None:
            component_name = self.get_unique_component_name()

        print("abc.add_component  component_name {}".format(component_name))
        node.add_component(model_type=model_type, name=component_name)

    def get_experiment_by_slice_id(self, slice_id):
        experiment = None
        try:
            experiment = list(
                filter(lambda x: x["slice_id"] == slice_id, self.experiments)
            )[0]
        except:
            print("Slice {} not found".format(slice_id))
        return experiment

    def get_experiment_by_name(self, slice_name):
        experiment = None
        try:
            experiment = list(
                filter(lambda x: x["slice_name"] == slice_name, self.experiments)
            )[0]
        except:
            print("Slice {} not found".format(slice_name))
        return experiment

    def pull_experiment_topology(self, experiment):
        # do not update if unsubmitted
        if experiment["editor_slice_state"] != self.EXPERIMENT_STATE_UNSUBMITTED:
            status, current_slice_topology = self.slice_manager.get_slice_topology(
                slice_object=experiment["slice"]
            )
            experiment["topology"] = current_slice_topology

    def update_current_experiment(self):
        update_experiment_topology(self.current_experiment)

    def update_all_experiment_topologies(self):
        for experiment in self.experiments:
            update_experiment_topology(experiment)

    def update_experiment_list(
        self, current_slice_name=None, excludes=[SliceState.Dead, SliceState.Closing]
    ) -> List[str]:
        """
        Update Slice list
        :param current_slice_name:
        :param excludes:
        :return:
        """
        print("XXXXXXXXXX update_experiment_list XXXXXXXXXX")

        # Get all existing slices from FABRIC slice_manager
        status, existing_slices = self.slice_manager.slices(excludes=excludes)

        # TODO: Remove this
        if status != Status.OK:
            print(
                "slice_manager.slices: Status: {}, Error: {}".format(
                    status, existing_slices
                )
            )
            existing_slices = []

        # Create new list of slices
        # new_experiments_list = []
        for slice in existing_slices:
            experiment = self.get_experiment_by_name(slice.name)

            if experiment == None:
                # New experiment
                print("Getting topology for new slice")
                experiment = {
                    "slice_name": slice.name,
                    "slice_id": slice.slice_id,
                    "editor_slice_state": self.EXPERIMENT_STATE_LIVE,
                    "slice": slice,
                    # ssh_keys
                    # detail_levels {'detail': HIGH, {'node1': LOW, 'node2': MED}}
                }

                status, experiment["topology"] = self.slice_manager.get_slice_topology(
                    slice_object=slice
                )
                if status != Status.OK:
                    print(
                        "Failed to get topology for slice found on FABRI: Status: {}, Error: {}".format(
                            status, experiment["topology"]
                        )
                    )

                self.experiments.append(experiment)
            else:
                # Existing experiment
                print("Getting topology for old slice")

                experiment["slice"] = slice
                # experiment['editor_slice_state'] = slice.slice_state
                status, experiment["topology"] = self.slice_manager.get_slice_topology(
                    slice_object=slice
                )
                if status != Status.OK:
                    print(
                        "Failed to get topology for existing slice: Status: {}, Error: {}".format(
                            status, experiment["topology"]
                        )
                    )

        # Mark submitted slices that are missing from FABRIC with error or Deleted
        status, all_slices = self.slice_manager.slices(includes=excludes)
        if status != Status.OK:
            print(
                "Failed to get all_slices: Status: {}, Error: {}".format(
                    status, all_slices
                )
            )

        for experiment in self.experiments:
            # Skip unsubmitted slices
            if experiment["editor_slice_state"] == self.EXPERIMENT_STATE_UNSUBMITTED:
                continue

            print("experiment: {} ".format(experiment))
            found_experiment = list(
                filter(lambda x: x.slice_id == experiment["slice_id"], all_slices)
            )
            if len(found_experiment) > 0:
                found_experiment = found_experiment[0]
            if found_experiment:
                print("found_experiment state: {}".format(found_experiment.slice_state))
                if found_experiment.slice_state == str(
                    SliceState.Dead
                ) or found_experiment.slice_state == str(SliceState.Closing):
                    # TODO need to deistinguish between error and deleted
                    experiment["editor_slice_state"] = self.EXPERIMENT_STATE_DELETED
                    experiment["slice"] = found_experiment
                else:
                    experiment["editor_slice_state"] = self.EXPERIMENT_STATE_LIVE
                    experiment["slice"] = found_experiment

    def submit_slice(self, experiment):
        slice_name = self.current_experiment["slice_name"]
        # ssh_key = self.current_experiment['ssh_key']
        slice_graph = self.current_experiment["topology"].serialize()

        # Request slice from Orchestrator
        status, reservations = self.slice_manager.create(
            slice_name=slice_name, slice_graph=slice_graph, ssh_key=self.ssh_key
        )
        print("Response Status {}".format(status))
        if status == Status.OK:
            print("Reservations created {}".format(reservations))
        else:
            print(f"Failure: {reservations}")

        self.current_experiment["slice_id"] = reservations[0].slice_id
        self.current_experiment["editor_slice_state"] = self.EXPERIMENT_STATE_LIVE

    @staticmethod
    def get_component_type_list():
        """
        Get component type list
        :return:
        """
        return_list = []
        for type in ComponentType:
            return_list.append(type.name)
        return return_list

    @staticmethod
    def get_component_model_list(type=None):
        """
        Get component model list
        :param type:
        :return:
        """
        if type is None:
            return ComponentCatalog.catalog_instance.copy()
        else:
            return list(
                filter(lambda x: x["Type"] == type, ComponentCatalog.catalog_instance)
            )

    @staticmethod
    def get_component_widget_options_list(type=None):
        """
        Get component widget options list
        :param type:
        :return:
        """
        return_list = []
        for component in AbcTopologyEditor.get_component_model_list(type):
            return_list.append(component["Model"] + " (" + component["Type"] + ")")

        return return_list

    def get_node_name_list(self):
        """
        Get node name list
        :return:
        """
        if self.current_experiment:
            return list(self.current_experiment["topology"].nodes)
        else:
            return []

    def get_unique_node_name(self):
        """
        Get Unique node name
        :return:
        """
        # need to go through all the nodes in the experiment
        num = 1
        name = "node" + str(num)
        while name in self.current_experiment["topology"].nodes:
            num += 1
            name = "node" + str(num)

        return name

    # def get_unique_component_name(self):
    #     """
    #     Get unique component name
    #     :return:
    #     """
    #     num = 1
    #     name = 'dev' + str(num)
    #     while len(list(filter(lambda x: x['component_name_widget'].value == name,
    #                           self.dashboards['node_dashboard']['components']))) > 0:
    #         num += 1
    #         name = 'dev' + str(num)
    #
    #     return name

    def get_unique_component_name(self, node=None):
        """
        Get unique component name
        :return:
        """
        if node == None:
            node = self.current_node

        num = 1
        name = "dev" + str(num)
        for component_name, component in node.components.items():
            if name == component_name:
                num += 1
                name = "node" + str(num)
            else:
                break

        return name

    @abstractmethod
    def start(self):
        """
        Start the editors
        :return:
        """



import socket
import time
import unittest

from fabrictestbed_extensions.fablib.fablib import FablibManager
from fabrictestbed_extensions.fablib.node import Node
from fabrictestbed_extensions.fablib.slice import Slice


class HelloFabricTests(unittest.TestCase):
    """
    Run some basic tests against the testbed.
    """

    def test_fablib_hello(self):
        """
        Create a slice with a single node, and echo a message from the node.
        """
        fablib = FablibManager()

        fablib.show_config()

        # fablib.list_sites()

        # Give the slice a unique name so that slice creation will not
        # fail (because there is an existing slice with the same name) and
        # we will have some hints about the test that created the slice.
        time_stamp = time.strftime("%Y-%m-%d %H:%M:%S")
        host = socket.gethostname()
        slice_name = f"integration test @ {time_stamp} on {host}"

        print(f"Creating slice '{slice_name}'..")
        slice = fablib.new_slice(name=slice_name)

        self.assertIsInstance(slice, Slice)

        try:
            # Add a node.
            node_name = "node-1"
            site_name = fablib.get_random_site()

            print(
                f"Adding node '{node_name}' at site '{site_name}' to slice '{slice_name}'.."
            )
            node = slice.add_node(name=node_name, site=site_name)

            self.assertIsInstance(node, Node)

            # Submit the slice.
            print(f"Submitting slice '{slice_name}'..")
            slice.submit()

            print(f"Slice '{slice_name}' status:")
            slice.show()

            print(f"Testing node '{node_name}' on slice '{slice_name}'...")
            for node in slice.get_nodes():
                stdout, stderr = node.execute(
                    "echo Hello, FABRIC from node `hostname -s`"
                )

                self.assertEqual(stdout, f"Hello, FABRIC from node {node_name}\n")
                self.assertEqual(stderr, "")

        finally:
            slice.delete()


if __name__ == "__main__":
    unittest.main()



import tempfile
import unittest

import paramiko

from fabrictestbed_extensions.fablib.fablib import FablibManager


class BastionHostTests(unittest.TestCase):
    """
    Test that bastion setup works.

    Fablib needs bastion host name, username, ssh key, and the
    (optional) key passphrase to function correctly.
    """

    def test_probe_bastion_host(self):
        """
        This should pass when fablib set up is correct.
        """
        fm = FablibManager(offline=True)
        self.assertTrue(fm.probe_bastion_host())

    def test_probe_bastion_host_no_username(self):
        """
        Test bastion with an empty username.
        """
        fm = FablibManager(offline=True, bastion_username="")
        self.assertEqual(fm.get_bastion_username(), "")
        self.assertRaises(paramiko.ssh_exception.SSHException, fm.probe_bastion_host)

    def test_probe_bastion_host_empty_key(self):
        """
        Test bastion with an empty key.
        """
        keyfile = tempfile.NamedTemporaryFile()

        fm = FablibManager(offline=True, bastion_key_location=keyfile.name)
        self.assertEqual(fm.get_bastion_key_location(), keyfile.name)
        self.assertRaises(paramiko.ssh_exception.SSHException, fm.probe_bastion_host)

    def test_probe_bastion_host_bad_key(self):
        """
        Test bastion with a key we just generated.
        """
        keyfile = tempfile.NamedTemporaryFile()

        rsa_key = paramiko.RSAKey.generate(bits=2048)
        rsa_key.write_private_key_file(keyfile.name)

        fm = FablibManager(offline=True, bastion_key_location=keyfile.name)
        self.assertEqual(fm.get_bastion_key_location(), keyfile.name)
        self.assertRaises(paramiko.ssh_exception.SSHException, fm.probe_bastion_host)



#!/usr/bin/env python3
#
# MIT License
#
# Copyright (c) 2023 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import socket
import time
import unittest
from ipaddress import IPv4Network

from fabrictestbed_extensions.fablib.fablib import fablib


class ModifyTests(unittest.TestCase):
    # Parts of this is extracted from the notebook
    # create_l2network_wide_area_auto.ipynb, while investigating
    # https://github.com/fabric-testbed/fabrictestbed-extensions/issues/261

    def setUp(self):
        # Create a slice
        time_stamp = time.strftime("%Y-%m-%d %H:%M:%S")
        host = socket.gethostname()
        slice_name = f"integration test @ {time_stamp} on {host}"
        self._slice = fablib.new_slice(name=slice_name)

    def tearDown(self):
        self._slice.delete()

    def test_modify_add_l2_l3_nodes(self):
        # Add nodes with L2 network, submit, add a third node with L3
        # network, add L3 network to the first two nodes, submit again.

        [site1, site2, site3] = fablib.get_random_sites(count=3)
        print(f"Sites: {site1}, {site2}, {site3}")

        print(f"Adding nodes to slice at {site1} and {site2}")
        self._add_l2(site1, site2)
        print(f"Submitting '{self._slice.get_name()}'")
        self._slice.submit()

        # Add a "marker" log line for ease of searching/spotting.
        logging.info("@@@@@@@@@@@@@@@ BEGIN MODIFY @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")

        print(f"Adding a third node at {site3}")
        self._add_l3(site1, site2, site3)
        print(f"Re-submitting '{self._slice.get_name()}'")
        self._slice.submit()

        logging.info("############### END MODIFY #################################")

        self._check_interfaces()

    def test_no_modify_l2_l3_nodes(self):
        # Add nodes with L2 network, add a third node, and L3 network,
        # then submit.

        print("Adding nodes, no modify")
        [site1, site2, site3] = fablib.get_random_sites(count=3)
        print(f"Sites: {site1}, {site2}, {site3}")

        print(f"Adding nodes and L2 network to slice at {site1} and {site2}")
        self._add_l2(site1, site2)

        print(f"Adding a node at {site3} and L3 network to slice at {site1}, {site2}")
        self._add_l3(site1, site2, site3)

        print("Submitting slice")
        self._slice.submit()

        self._check_interfaces()

    def _add_l2(self, site1, site2):
        """
        Make a slice with an L2 network and two nodes.
        """
        # Add an L2 network.
        network_name = "net-L2"
        l2_net = self._slice.add_l2network(
            name=network_name, subnet=IPv4Network("192.168.1.0/24")
        )

        # Add some nodes with L2 networking.
        self.__add_node_l2(site1, l2_net)
        self.__add_node_l2(site2, l2_net)

    def __add_node_l2(self, site, net):
        # Set up a node with a NIC.
        node_name = f"node-{site}"
        print(f"Adding node {node_name}")
        node = self._slice.add_node(name=node_name, site=site)

        ifname = f"nic-L2-{site}"
        print(f"Adding {ifname} to {node_name}")

        iface = node.add_component(model="NIC_Basic", name=ifname).get_interfaces()[0]
        iface.set_mode("auto")

        net.add_interface(iface)

    def _add_l3(self, site1, site2, site3):
        """
        Add an L3 network to the slice.
        """
        # Add L3 networking to existing nodes.
        self.__add_l3_to_node(site1)
        self.__add_l3_to_node(site2)

        # Add another L3 network tied to a new node.
        self.__add_l3_node(site3)

    def __add_l3_to_node(self, site):
        node_name = f"node-{site}"
        node = self._slice.get_node(name=node_name)
        self.__add_l3_iface(site, node)

    def __add_l3_node(self, site):
        node_name = f"node-L3-{site}"
        print(f"Adding node {node_name}")
        node = self._slice.add_node(name=node_name, site=site)

        self.__add_l3_iface(site, node)

    def __add_l3_iface(self, site, node):
        if_name = f"nic-L3-{site}"
        print(f"Adding {if_name} to {node.get_name()}")

        iface = node.add_component(model="NIC_Basic", name=if_name).get_interfaces()[0]
        iface.set_mode("auto")

        l3_net_name = f"net-L3-{site}"
        print(f"Adding L3 network {l3_net_name}, with {if_name}")
        l3_net = self._slice.add_l3network(name=l3_net_name, type="IPv4")

        l3_net.add_interface(iface)

    def _check_interfaces(self):
        print("============ interfaces ====================================")

        ifaces = self._slice.get_interfaces()
        for iface in ifaces:
            print(f"{iface}")

        print("============ networks ======================================")

        networks = self._slice.get_networks()
        for network in networks:
            print(f"{network}")

        print("============================================================")

        errors = []

        for iface in ifaces:
            if iface.get_ip_addr() is None:
                errors.append(f"iface {iface.get_name()} has no IP address")

        self.assertEqual(errors, [])



import os
import pathlib
import tempfile
import unittest

from fabrictestbed.fabric_manager import FabricManagerException
from fabrictestbed.token_manager.token_manager import TokenManagerException
from fabrictestbed.util.constants import Constants

from fabrictestbed_extensions.fablib.config.config import Config, ConfigException
from fabrictestbed_extensions.fablib.constants import Constants as FablibConstants
from fabrictestbed_extensions.fablib.fablib import FablibManager


class FablibManagerTests(unittest.TestCase):
    """
    Some basic tests to exercise FablibManager.

    Instantiating FablibManager will throw some exceptions if some
    required environment variables are not set.
    """

    required_env_vars = [
        Constants.FABRIC_CREDMGR_HOST,
        Constants.FABRIC_ORCHESTRATOR_HOST,
        Constants.FABRIC_PROJECT_ID,
        # Constants.FABRIC_TOKEN_LOCATION,
        FablibConstants.FABRIC_BASTION_HOST,
        FablibConstants.FABRIC_BASTION_USERNAME,
        FablibConstants.FABRIC_BASTION_KEY_LOCATION,
    ]

    DUMMY_TOKEN_LOCATION = str(
        pathlib.Path(__file__).parent / "data" / "dummy-token.json"
    )
    FABRIC_RC_LOCATION = str(pathlib.Path(__file__).parent / "data" / "dummy_fabric_rc")

    def setUp(self):
        # Run each test with an empty environment.
        os.environ.clear()

        # Create an empty configuration file, so that we will be
        # really testing with a clean slate.  Creating the
        # configuration file in read-only mode should ensure that it
        # will remain empty.
        self.rcfile = tempfile.NamedTemporaryFile(mode="r")
        self.rcfile.flush()

    def tearDown(self):
        self.rcfile.close()

    def test_fablib_manager_test_only_cm_host(self):
        os.environ[Constants.FABRIC_CREDMGR_HOST] = "dummy"
        os.environ[Constants.FABRIC_TOKEN_LOCATION] = self.DUMMY_TOKEN_LOCATION
        self.assertRaises(ConnectionError, FablibManager, fabric_rc=self.rcfile.name)

    def test_fablib_manager_test_only_orchestrator_host(self):
        os.environ[Constants.FABRIC_ORCHESTRATOR_HOST] = "dummy"
        os.environ[Constants.FABRIC_TOKEN_LOCATION] = self.DUMMY_TOKEN_LOCATION
        self.assertRaises(ConnectionError, FablibManager, fabric_rc=self.rcfile.name)

    def test_fablib_manager_test_only_project_id(self):
        os.environ[Constants.FABRIC_PROJECT_ID] = "dummy"
        os.environ[Constants.FABRIC_TOKEN_LOCATION] = self.DUMMY_TOKEN_LOCATION
        self.assertRaises(Exception, FablibManager, fabric_rc=self.rcfile.name)

    def test_fablib_manager_test_only_token_location(self):
        os.environ[Constants.FABRIC_TOKEN_LOCATION] = "dummy"
        self.assertRaises(ConfigException, FablibManager, fabric_rc=self.rcfile.name)

    def test_fablib_manager_test_with_no_token_file(self):
        # Should fail when token location is not a valid path.

        # set all required env vars.
        os.environ[Constants.FABRIC_PROJECT_ID] = "dummy_project_id"
        os.environ[FablibConstants.FABRIC_BASTION_HOST] = "dummy_bastion_host"
        os.environ[FablibConstants.FABRIC_BASTION_USERNAME] = "dummy_bastion_user_name"
        os.environ[Constants.FABRIC_TOKEN_LOCATION] = "dummy"

        # FablibManager() without a valid token or token location
        # should raise a "SliceManagerException: Unable to refresh tokens: no refresh token found!
        self.assertRaises(ConfigException, FablibManager, fabric_rc=self.rcfile.name)

    def test_fablib_manager_test_with_dummy_token(self):
        # TODO: That FablibManager() calls build_slice_manager()
        # complicates writing a test for it.  It eventually makes a
        # network call to credential manager API, but it is not right
        # for a unit test to do such a thing.  We could probably
        # somehow mock a CM here?

        # set all required env vars.
        for var in self.required_env_vars:
            os.environ[var] = "dummy"

        # '.invalid' is an invalid host per RFC 6761, so this test
        # must fail without ever making a successful network call.
        os.environ[Constants.FABRIC_CREDMGR_HOST] = ".invalid"
        os.environ[Constants.FABRIC_TOKEN_LOCATION] = self.DUMMY_TOKEN_LOCATION

        self.assertRaises(ValueError, FablibManager, fabric_rc=self.rcfile.name)

    def test_fablib_manager_with_empty_config(self):
        # Check that an empty configuration file will cause
        # FablibManager to raise an error.
        rcfile = tempfile.NamedTemporaryFile()
        rcfile.flush()
        os.environ[Constants.FABRIC_TOKEN_LOCATION] = self.DUMMY_TOKEN_LOCATION

        with self.assertRaises(Exception):
            FablibManager(fabric_rc=rcfile.name)

    def test_fablib_manager_with_some_config(self):
        # Test with some configuration in the rc file.
        rcfile = tempfile.NamedTemporaryFile()

        # Write all configuration except FABRIC_TOKEN_LOCATION.  The
        # token location is a special case because when it is set,
        # Fablib will need network access to reach CredentialManager.
        for var in self.required_env_vars:
            rcfile.write(f"export {var} = dummy\n".encode())

        rcfile.flush()

        with self.assertRaises(ConfigException) as ctx:
            FablibManager(fabric_rc=rcfile.name)

        # Check that the error is what we expected.
        self.assertIsInstance(ctx.exception, ConfigException)

        # Check that the error message is what we expected: the only
        # error should be about missing token.
        self.assertTrue(
            "Token file does not exist, please provide the token at location"
            in str(ctx.exception)
        )

    def test_FablibManager_no_config_no_env_var(self):
        # Instantiate Fablib manager without any config or environment variables
        # Results in using the defaults and initialization fails when token file is not found!
        self.assertRaises(
            ConfigException,
            FablibManager,
        )

    def test_FablibManager_no_config_no_env_var_token_location(self):
        # Instantiate Fablib manager without any config or environment variables

        with self.assertRaises(Exception) as ctx:
            FablibManager(
                token_location=self.DUMMY_TOKEN_LOCATION,
                fabric_rc=self.FABRIC_RC_LOCATION,
            )

        # Check that the error is what we expected.
        self.assertIsInstance(ctx.exception, Exception)

        # Check that the error message is what we expected: the only
        # error should be about missing token.
        self.assertEqual(
            "Unable to validate provided token: ValidateCode.UNPARSABLE_TOKEN/Not enough segments",
            str(ctx.exception),
        )

    def test_FablibManager_no_config_no_env_var_token_location_offline(self):
        # Instantiate Fablib manager without any config or environment variables

        with self.assertRaises(AttributeError) as ctx:
            FablibManager(
                token_location=self.DUMMY_TOKEN_LOCATION,
                offline=True,
                fabric_rc=self.FABRIC_RC_LOCATION,
            )

        # Check that the error is what we expected.
        self.assertIsInstance(ctx.exception, AttributeError)

    def test_FablibManager_no_config_no_env_var_token_location_offline_project_id_bastion_user_name(
        self,
    ):
        # Instantiate Fablib manager without any config or environment variables
        project_id = "DUMMY_PROJECT_ID"
        bastion_username = "DUMMY_BASTION_USERNAME"

        fablib = FablibManager(
            token_location=self.DUMMY_TOKEN_LOCATION,
            offline=True,
            project_id=project_id,
            bastion_username=bastion_username,
            fabric_rc=self.FABRIC_RC_LOCATION,
        )
        self.assertEqual(project_id, fablib.get_project_id())
        self.assertEqual(bastion_username, fablib.get_bastion_username())
        self.assertEqual(self.DUMMY_TOKEN_LOCATION, fablib.get_token_location())

        for attrs, attr_props in Config.REQUIRED_ATTRS.items():
            if attrs not in [
                FablibConstants.PROJECT_ID,
                FablibConstants.BASTION_USERNAME,
                FablibConstants.TOKEN_LOCATION,
            ]:
                default_value = attr_props.get(FablibConstants.DEFAULT)
                if default_value:
                    self.assertEqual(default_value, fablib.runtime_config.get(attrs))

    def test_FablibManager_no_config_no_env_var_token_location_offline_project_id_bastion_user_name_env(
        self,
    ):
        # Instantiate Fablib manager without any config or environment variables
        project_id = "DUMMY_PROJECT_ID"
        bastion_username = "DUMMY_BASTION_USERNAME"
        os.environ[Constants.FABRIC_TOKEN_LOCATION] = self.DUMMY_TOKEN_LOCATION
        os.environ[Constants.FABRIC_PROJECT_ID] = project_id
        os.environ[FablibConstants.FABRIC_BASTION_USERNAME] = bastion_username

        fablib = FablibManager(fabric_rc=self.FABRIC_RC_LOCATION, offline=True)
        self.assertEqual(project_id, fablib.get_project_id())
        self.assertEqual(bastion_username, fablib.get_bastion_username())
        self.assertEqual(self.DUMMY_TOKEN_LOCATION, fablib.get_token_location())

        for attrs, attr_props in Config.REQUIRED_ATTRS.items():
            if attrs not in [
                FablibConstants.PROJECT_ID,
                FablibConstants.BASTION_USERNAME,
                FablibConstants.TOKEN_LOCATION,
            ]:
                default_value = attr_props.get(FablibConstants.DEFAULT)
                if default_value:
                    self.assertEqual(default_value, fablib.runtime_config.get(attrs))



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

"""
Methods to work with FABRIC components_.

.. _components: https://learn.fabric-testbed.net/knowledge-base/glossary/#component

You normally would not create :class:`Component` objects directly with
a constructor call; they are created when you invoke
:py:func:`fabrictestbed_extensions.fablib.node.Node.add_component()`,
like so::

    node.add_component(model='NVME_P4510', name="nvme1")
    node.add_component(model='NIC_Basic', name="nic1")
"""

from __future__ import annotations

import json
from typing import TYPE_CHECKING

import jinja2

from fabrictestbed_extensions.fablib.constants import Constants

if TYPE_CHECKING:
    from fabrictestbed_extensions.fablib.slice import Slice
    from fabrictestbed_extensions.fablib.node import Node
    from fabrictestbed_extensions.fablib.interface import Interface
    from fabrictestbed_extensions.fablib.interface import Interface

import logging
from typing import List

from fabrictestbed.slice_editor import Component as FimComponent
from fabrictestbed.slice_editor import ComponentModelType, Flags, Labels, UserData
from tabulate import tabulate


class Component:
    """
    A class for working with FABRIC components.
    """

    component_model_map = {
        Constants.CMP_NIC_Basic: ComponentModelType.SharedNIC_ConnectX_6,
        Constants.CMP_NIC_ConnectX_6: ComponentModelType.SmartNIC_ConnectX_6,
        Constants.CMP_NIC_ConnectX_5: ComponentModelType.SmartNIC_ConnectX_5,
        Constants.CMP_NIC_P4: Constants.P4_DedicatedPort,
        Constants.CMP_NVME_P4510: ComponentModelType.NVME_P4510,
        Constants.CMP_GPU_TeslaT4: ComponentModelType.GPU_Tesla_T4,
        Constants.CMP_GPU_RTX6000: ComponentModelType.GPU_RTX6000,
        Constants.CMP_GPU_A40: ComponentModelType.GPU_A40,
        Constants.CMP_GPU_A30: ComponentModelType.GPU_A30,
        Constants.CMP_NIC_OpenStack: ComponentModelType.SharedNIC_OpenStack_vNIC,
        Constants.CMP_FPGA_Xilinx_U280: ComponentModelType.FPGA_Xilinx_U280,
        Constants.CMP_FPGA_Xilinx_SN1022: ComponentModelType.FPGA_Xilinx_SN1022,
    }

    def __str__(self):
        """
        Creates a tabulated string describing the properties of the component.

        Intended for printing component information.

        :return: Tabulated string of component information
        :rtype: String
        """
        table = [
            ["Name", self.get_name()],
            ["Details", self.get_details()],
            ["Disk", self.get_disk()],
            ["Units", self.get_unit()],
            ["PCI Address", self.get_pci_addr()],
            ["Model", self.get_model()],
            ["Type", self.get_type()],
        ]

        return tabulate(table)

    def get_fablib_manager(self):
        """
        Get the Fabric library manager associated with the component.
        """
        return self.get_slice().get_fablib_manager()

    def toJson(self):
        """
        Returns the component attributes as a json string

        :return: slice attributes as json string
        :rtype: str
        """
        return json.dumps(self.toDict(), indent=4)

    @staticmethod
    def get_pretty_name_dict():
        """
        Returns the mapping used when rendering table headers.
        """
        return {
            "name": "Name",
            "short_name": "Short Name",
            "details": "Details",
            "disk": "Disk",
            "units": "Units",
            "pci_address": "PCI Address",
            "model": "Model",
            "type": "Type",
            "dev": "Device",
            "node": "Node",
            "numa": "Numa Node",
        }

    def toDict(self, skip=[]):
        """
        Returns the component attributes as a dictionary

        :return: slice attributes as dictionary
        :rtype: dict
        """
        return {
            "name": str(self.get_name()),
            "short_name": str(self.get_short_name()),
            "details": str(self.get_details()),
            "disk": str(self.get_disk()),
            "units": str(self.get_unit()),
            "pci_address": str(self.get_pci_addr()),
            "model": str(self.get_model()),
            "type": str(self.get_type()),
            "dev": str(self.get_device_name()),
            "node": str(self.get_node().get_name()),
            "numa": str(self.get_numa_node()),
        }

    def generate_template_context(self):
        context = self.toDict()
        context["interfaces"] = []
        # for interface in self.get_interfaces():
        #    context["interfaces"].append(interface.get_name())

        #    context["interfaces"].append(interface.generate_template_context())
        return context

    def get_template_context(self):
        return self.get_slice().get_template_context(self)

    def render_template(self, input_string):
        environment = jinja2.Environment()
        # environment.json_encoder = json.JSONEncoder(ensure_ascii=False)
        template = environment.from_string(input_string)
        output_string = template.render(self.get_template_context())

        return output_string

    def show(
        self, fields=None, output=None, quiet=False, colors=False, pretty_names=True
    ):
        """
        Show a table containing the current component attributes.

        There are several output options: "text", "pandas", and "json" that determine the format of the
        output that is returned and (optionally) displayed/printed.

        output:  'text': string formatted with tabular
                  'pandas': pandas dataframe
                  'json': string in json format

        fields: json output will include all available fields.

        Example: fields=['Name','PCI Address']

        :param output: output format
        :type output: str
        :param fields: list of fields to show
        :type fields: List[str]
        :param quiet: True to specify printing/display
        :type quiet: bool
        :param colors: True to specify state colors for pandas output
        :type colors: bool
        :return: table in format specified by output parameter
        :rtype: Object
        """
        data = self.toDict()

        # fields = ["Name", "Details", "Disk", "Units", "PCI Address",
        #        "Model", "Type"
        #         ]

        if pretty_names:
            pretty_names_dict = self.get_pretty_name_dict()
        else:
            pretty_names_dict = {}

        table = self.get_fablib_manager().show_table(
            data,
            fields=fields,
            title="Component",
            output=output,
            quiet=quiet,
            pretty_names_dict=pretty_names_dict,
        )

        return table

    def list_interfaces(
        self, fields=None, output=None, quiet=False, filter_function=None
    ):
        """
        Lists all the interfaces in the component with their attributes.

        There are several output options: "text", "pandas", and "json" that determine the format of the
        output that is returned and (optionally) displayed/printed.

        output:  'text': string formatted with tabular
                  'pandas': pandas dataframe
                  'json': string in json format

        fields: json output will include all available fields/columns.

        Example: fields=['Name','MAC']

        filter_function:  A lambda function to filter data by field values.

        Example: filter_function=lambda s: s['Node'] == 'Node1'

        :param output: output format
        :type output: str
        :param fields: list of fields (table columns) to show
        :type fields: List[str]
        :param quiet: True to specify printing/display
        :type quiet: bool
        :param filter_function: lambda function
        :type filter_function: lambda
        :return: table in format specified by output parameter
        :rtype: Object
        """

        ifaces = []
        for iface in self.get_interfaces():
            ifaces.append(iface.get_name())

        name_filter = lambda s: s["Name"] in set(ifaces)
        if filter_function != None:
            filter_function = lambda x: filter_function(x) + name_filter(x)
        else:
            filter_function = name_filter

        return self.get_slice().list_interfaces(
            fields=fields, output=output, quiet=quiet, filter_function=filter_function
        )

    @staticmethod
    def calculate_name(node: Node = None, name: str = None) -> str:
        """
        Not intended for API use
        """
        # Hack to make it possile to find interfaces
        return f"{node.get_name()}-{name}"

    @staticmethod
    def new_component(
        node: Node = None, model: str = None, name: str = None, user_data: dict = {}
    ):
        """
        Not intended for API use

        Creates a new FIM component on the fablib node inputted.

        :param node: the fablib node to build the component on
        :type node: Node
        :param model: the name of the component type to build
        :type model: str
        :param name: the name of the new component
        :type name: str
        :return: the new fablib compoent
        :rtype: Component
        """
        # Hack to make it possile to find interfaces
        name = Component.calculate_name(node=node, name=name)

        component = Component(
            node=node,
            fim_component=node.fim_node.add_component(
                model_type=Component.component_model_map[model], name=name
            ),
        )
        component.set_user_data(user_data)
        return component

    def __init__(self, node: Node = None, fim_component: FimComponent = None):
        """
        Typically invoked when you add a component to a ``Node``.

        .. note ::

            ``Component`` constructer is not meant to be directly used.

        :param node: the fablib node to build the component on
        :type node: Node

        :param fim_component: the FIM component this object represents
        :type fim_component: FIMComponent
        """
        super().__init__()
        self.fim_component = fim_component
        self.node = node
        self.interfaces = None

    def get_interfaces(self, include_subs: bool = True) -> List[Interface]:
        """
        Gets the interfaces attached to this fablib component's FABRIC component.

        :param include_subs: Flag indicating if sub interfaces should be included
        :type include_subs: bool

        :return: a list of the interfaces on this component.
        :rtype: List[Interface]
        """

        from fabrictestbed_extensions.fablib.interface import Interface

        if not self.interfaces:
            self.interfaces = []
            for fim_interface in self.get_fim_component().interface_list:
                iface = Interface(component=self, fim_interface=fim_interface)
                self.interfaces.append(iface)
                if include_subs:
                    child_interfaces = iface.get_interfaces()
                    if child_interfaces and len(child_interfaces):
                        self.interfaces.extend(child_interfaces)

        return self.interfaces

    def get_fim_component(self) -> FimComponent:
        """
        Not recommended for most users.

        GGets the FABRIC component this fablib component represents. This method
        is used to access data at a lower level than FABlib.

        :return: the FABRIC component on this component
        :rtype: FIMComponent
        """
        return self.fim_component

    def get_slice(self) -> Slice:
        """
        Gets the fablib slice associated with this component's node.

        :return: the slice this component is on
        :rtype: Slice
        """
        return self.node.get_slice()

    def get_node(self) -> Node:
        """
        Gets the fablib node this component is associated with.

        :return: the node this component is on
        :rtype: Node
        """
        return self.node

    def get_site(self) -> str:
        """
        Gets the name of the site this component's node is on.

        :return: the site name this node is on
        :rtype: String
        """
        return self.node.get_site()

    def get_short_name(self):
        """
        Gets the short name of the component.
        """
        # strip of the extra parts of the name added by fim
        return self.get_name()[len(f"{self.get_node().get_name()}-") :]

    def get_name(self) -> str:
        """
        Gets the name of this component from the FABRIC component.

        :return: the name of this component
        :rtype: str
        """
        return self.get_fim_component().name

    def get_details(self) -> str:
        """
        Not intended for API use
        """
        return self.get_fim_component().details

    def get_numa_node(self) -> str:
        """
        Get the Numa Node assigned to the device
        """
        try:
            numa = self.get_fim_component().get_property(pname="label_allocations").numa
            if numa is not None:
                if isinstance(numa, str):
                    return numa
                if isinstance(numa, list):
                    return numa[0]
        except Exception as e:
            logging.error(f"get_numa_node failed: {e}")
            return None

    def get_disk(self) -> int:
        """
        Gets the amount of disk space on this component.

        :return: this component's disk space
        :rtype: int
        """
        return self.get_fim_component().get_property(pname="capacity_allocations").disk

    def get_unit(self) -> int:
        """
        Get unit count for this component.

        :return: unit
        :rtype: int
        """
        return self.get_fim_component().get_property(pname="capacity_allocations").unit

    def get_pci_addr(self) -> str:
        """
        Get the PIC device ID for this component.

        :return: PCI device ID
        :rtype: String
        """
        return self.get_fim_component().get_property(pname="label_allocations").bdf

    def get_model(self) -> str:
        """
        Get FABlib model name for this component.

        :return: FABlib model name
        :rtype: String
        """
        # TODO: This a hack that need a real fix
        if (
            str(self.get_type()) == "SmartNIC"
            and str(self.get_fim_model()) == "ConnectX-6"
        ):
            return Constants.CMP_NIC_ConnectX_6
        elif (
            str(self.get_type()) == "SmartNIC"
            and str(self.get_fim_model()) == "ConnectX-5"
        ):
            return Constants.CMP_NIC_ConnectX_5
        elif str(self.get_type()) == "NVME" and str(self.get_fim_model()) == "P4510":
            return Constants.CMP_NVME_P4510
        elif str(self.get_type()) == "GPU" and str(self.get_fim_model()) == "Tesla T4":
            return Constants.CMP_GPU_TeslaT4
        elif str(self.get_type()) == "GPU" and str(self.get_fim_model()) == "RTX6000":
            return Constants.CMP_GPU_RTX6000
        elif (
            str(self.get_type()) == "SharedNIC"
            and str(self.get_fim_model()) == "ConnectX-6"
        ):
            return Constants.CMP_NIC_Basic

    def get_reservation_id(self) -> str or None:
        """
        Get reservation ID for this component.

        :return:  reservation ID
        :rtype: String
        """
        try:
            # This does not work
            # print(f"{self.get_fim_component()}")
            return (
                self.get_fim_component()
                .get_property(pname="reservation_info")
                .reservation_id
            )
        except:
            return None

    def get_reservation_state(self) -> str or None:
        """
        Get reservation state for this component.

        :return:  reservation state
        :rtype: String
        """
        try:
            return (
                self.get_fim_component()
                .get_property(pname="reservation_info")
                .reservation_state
            )
        except:
            return None

    def get_error_message(self) -> str:
        """
        Get error message for this component.

        :return:  reservation state
        :rtype: String
        """
        try:
            return (
                self.get_fim_component()
                .get_property(pname="reservation_info")
                .error_message
            )
        except:
            return ""

    def get_fim_model(self) -> str:
        """
        Not for API use
        """
        return self.get_fim_component().model

    def get_type(self) -> str:
        """
        Not for API use

        Gets the type of this component.

        :return: the type of component
        :rtype: str
        """
        return self.get_fim_component().type

    def configure_nvme(self, mount_point=""):
        """
        Configure the NVMe drive.

        Note this works but may be reorganized.

        :param mount_point: The mount point in the filesystem. Default = "" later reassigned to /mnt/{linux device name}
        :type mount_point: String
        """
        output = []
        try:
            device_pci_id = self.get_pci_addr()[0]
            stdout, stderr = self.node.execute(
                f'basename `sudo ls -l /sys/block/nvme*|grep "'
                f"{device_pci_id}\"|awk '{{print $9}}'`",
                quiet=True,
            )
            if stderr != "":
                output.append(
                    f"Cannot find NVME device name for PCI ID : {device_pci_id}"
                )
                raise Exception
            device_name = stdout.strip()
            block_device_name = f"/dev/{device_name}"
            output.append(self.node.execute(f"sudo fdisk -l {block_device_name}"))
            output.append(
                self.node.execute(f"sudo parted -s {block_device_name} mklabel gpt")
            )
            output.append(
                self.node.execute(f"sudo parted -s {block_device_name} print")
            )
            output.append(
                self.node.execute(
                    f"sudo parted -s {block_device_name} print unit MB print free"
                )
            )
            output.append(
                self.node.execute(
                    f"sudo parted -s --align optimal "
                    f"{block_device_name} "
                    f"mkpart primary ext4 0% 100%"
                )
            )
            output.append(self.node.execute(f"lsblk {block_device_name}"))
            output.append(self.node.execute(f"sudo mkfs.ext4 {block_device_name}p1"))
            # This is to use a unique mountpoint when it is not provided by the user
            if mount_point == "":
                mount_point = f"/mnt/{device_name}"
            output.append(
                self.node.execute(
                    f"sudo mkdir -p "
                    f"{mount_point}"
                    f" && sudo mount "
                    f"{block_device_name}"
                    f"p1 "
                    f"{mount_point}"
                )
            )
            output.append(self.node.execute(f"df -h {mount_point}"))
        except Exception as e:
            logging.error(f"config_nvme Fail: {self.get_name()}:", exc_info=True)
            raise Exception(str(output))

        return output

    def get_device_name(self) -> str:
        """
        Not for API use
        """
        labels = self.get_fim_component().get_property(pname="label_allocations")
        return labels.device_name

    @staticmethod
    def new_storage(node: Node, name: str, auto_mount: bool = False):
        """
        Not intended for API use

        Creates a new FIM component on the fablib node inputted.

        :param node: the fablib node to build the component on
        :param name: the name of the new component
        :param auto_mount: True - mount the storage; False - do not mount
        :return: the new fablib compoent
        :rtype: Component
        """
        # Hack to make it possile to find interfaces

        fim_component = node.fim_node.add_storage(
            name=name,
            labels=Labels(local_name=name),
            flags=Flags(auto_mount=auto_mount),
        )
        return Component(node=node, fim_component=fim_component)

    def get_fim(self):
        """
        Gets the component's FABRIC Information Model (fim) object.

        This method is used to access data at a lower level than
        FABlib.
        """
        return self.get_fim_component()

    def set_user_data(self, user_data: dict):
        """
        Set the user data for the component.

        This method stores the given user data dictionary as a JSON
        string in the FIM object associated with the component.

        :param user_data: The user data to be set.
        :type user_data: dict
        """
        self.get_fim().set_property(
            pname="user_data", pval=UserData(json.dumps(user_data))
        )

    def get_user_data(self) -> dict:
        """
        Retrieve the user data for the component.

        This method fetches the user data stored in the FIM object
        associated with the component and returns it as a dictionary.
        If an error occurs, it returns an empty dictionary.

        :return: The user data dictionary.
        :rtype: dict
        """
        try:
            return json.loads(str(self.get_fim().get_property(pname="user_data")))
        except:
            return {}

    def delete(self):
        """
        Remove the component from the slice/node.
        """
        if self.get_interfaces():
            for interface in self.get_interfaces():
                interface.delete()

        self.get_slice().get_fim_topology().nodes[
            self.get_node().get_name()
        ].remove_component(name=self.get_name())
